#pragma once
#include <ts/scenario/functors/scenario.hpp>
#include <ts/scenario/functors/methods.hpp>
#include <ts/scenario/scenario.hpp>
#include <ts/scenario/signatures.hpp>

#include <vector>

namespace cpptesk
{
namespace ts
{

/// @brief This class provides utility methods to construct test scenarios for CTESK
///
class ScenarioBase
{
public:
    /// @brief This class specifies placeholders for ScenarioBase::setup methods.
    /// @detailed The placeholders defined in this class mean that <tt>setup</tt> methods should
    /// use corresponding virtual method instead of the placeholder.
    ///
    /// <p>Example</p>
    /// Code snippet <tt>setup("MyScenario", UseVirtual::init, UseVirtual::finish)</tt> instructs
    /// that the test scenario should use virtual <tt>init</tt> and <tt>finish</tt> methods.
    struct UseVirtual
    {
        /// Placeholder for virtual <tt>init</tt> method
        static const InitMethod<UseVirtual>::type init;
        /// Placeholder for virtual <tt>fihish</tt> method
        static const FinishMethod<UseVirtual>::type finish;
        /// Placeholder for virtual <tt>get_state</tt> method
        static const GetStateMethod<UseVirtual,Object*>::type get_state;
        /// Placeholder for virtual <tt>get_state_id</tt> method
        static const GetStateIdMethod<UseVirtual>::type  get_state_id;
        /// Placeholder for virtual <tt>observe_model_state</tt> method
        static const ObserveModelStateMethod<UseVirtual>::type observe_model_state;
        /// Placeholder for virtual <tt>is_stationary_model_state</tt> method
        static const IsStationaryStateMethod<UseVirtual>::type is_stationary_model_state;
        /// Placeholder for virtual <tt>save_model_state</tt> method
        static const SaveModelStateMethod<UseVirtual, Object*>::type save_model_state;
        /// Placeholder for virtual <tt>restore_model_state</tt> method
        static const RestoreModelStateMethod<UseVirtual>::type restore_model_state;
    };
    /// @brief This class specifies placeholders for ScenarioBase::setup methods.
    /// @detailed The placeholders defined in this class mean that <tt>setup</tt> methods should
    /// not use any method/functor instead of the placeholder.
    ///
    /// <p>Example</p>
    /// Code snippet <tt>setup("MyScenario", UseVirtual::init, NoMethod::finish)</tt> instructs
    /// that the test scenario should use virtual <tt>init</tt> method and has no <tt>finish</tt> method.
    struct NoMethod
    {
        /// Placeholder specifies that the test scenario has no <tt>init</tt>
        static const InitMethod<NoMethod>::type init;
        /// Placeholder specifies that the test scenario has no <tt>finish</tt>
        static const FinishMethod<NoMethod>::type finish;
        /// Placeholder specifies that the test scenario has no <tt>get_state</tt>
        static const GetStateMethod<NoMethod,Object*>::type get_state;
        /// Placeholder specifies that the test scenario has no <tt>get_state_id</tt>
        static const GetStateIdMethod<NoMethod>::type  get_state_id;
        /// Placeholder specifies that the test scenario has no <tt>observe_model_state</tt>
        static const ObserveModelStateMethod<NoMethod>::type observe_model_state;
        /// Placeholder specifies that the test scenario has no <tt>is_stationary_model_state</tt>
        static const IsStationaryStateMethod<NoMethod>::type is_stationary_model_state;
        /// Placeholder specifies that the test scenario has no <tt>save_model_state</tt> method
        static const SaveModelStateMethod<NoMethod, Object*>::type save_model_state;
        /// Placeholder specifies that the test scenario has no <tt>restore_model_state</tt>
        static const RestoreModelStateMethod<NoMethod>::type restore_model_state;
    };
public:
    /// @brief Initialization method of the test scenario.
    /// @detailed This method will be called by the test engine if <tt>setup</tt> method is called with
    /// <tt>UseVirtual::init</tt> placeholder.
    /// @param argc number of arguments in @a argv array.
    /// @param argv array of command-line parameters submitted to test engine's <tt>start</tt> function.
    /// @return <tt>true</tt> if initialization was successful and false otherwise.
    /// @note The method in the base class does nothing and returns <tt>true</tt>..
    virtual bool init(int argc, char ** argv);
    /// @brief Finalization method of the test scenario.
    /// @detailed This method will be called by the test engine if <tt>setup</tt> method is called with
    /// <tt>UseVirtual::finish</tt> placeholder.
    /// @note The method in the base class does nothing.
    virtual void finish();

    /// @brief GetState method of the test scenario.
    /// @detailed This method will be called by the test engine if <tt>setup</tt> method is called with
    /// <tt>UseVirtual::get_state</tt> placeholder.
    /// @return <tt>true</tt> if initialization was successful and false otherwise.
    /// @note The method in the base class does nothing and returns <tt>true</tt>..
    virtual Object* get_state_object();
    virtual String* get_state_id(Object* obj);

    virtual bool is_stationary_model_state();
    virtual void restore_model_state(Object *);
    virtual Object* save_model_state_object();
    virtual void observe_model_state();

    typedef std::vector<cpptesk::ts::scenario_method_base*> Actions;

protected:
    CppTestScenario _dfsm;

    bool _scenario_valid;
    bool _scenario_async;

    const char * _name;
    Actions _scenario_methods;

    scenario_init_base* _init;
    scenario_finish_base* _finish;
    scenario_get_state_base* _get_state;
    scenario_get_state_id_base * _get_state_id;
    scenario_is_stationary_state_base* _is_stationary;
    scenario_observe_state_base * _observe_state;
    scenario_restore_model_state_base* _restore_state;
    scenario_save_model_state_base * _save_model_state;

    void set_default_init();
    void set_default_finish();
    void set_default_get_state();
    void set_default_get_state_id();
    void set_default_is_stationary_state();
    void set_default_save_model_state();
    void set_default_restore_model_state();
    void set_default_observe_state();

    void remove_init();
    void remove_finish();
    void remove_get_state();
    void remove_get_state_id();
    void remove_is_stationary_state();
    void remove_save_model_state();
    void remove_restore_model_state();
    void remove_observe_state();

    ScenarioBase();
    ScenarioBase(const char * name);

    inline void invalidate()
    {
        _scenario_valid = false;
    }

    /** \brief    .
     *
     * \param name const char*    .
     * \return const char*   .
     *
     */
    const char * set_name(const char * name);
    /** \brief   .
     *    \a func    scenario_method<S, T>.
     *
     * \param name   .
     * \param func    .
     *
     * \tparam S   .
     * \tparam T  ,    .
     */
    template<typename S, typename T>
    void add_scenario_method(const char * name, bool (S::*func)(T&))
    {
        invalidate();
        scenario_method_base* method = new scenario_method<S, T>(name, *dynamic_cast<S*>(this), func);
        _scenario_methods.push_back(method);
    }

    /** \brief   .
     *
     * \param method     .
     */
    void add_scenario_method(scenario_method_base* method)
    {
        invalidate();
        _scenario_methods.push_back(method);
    }

    /** \brief Remove all registered scenatio methods
     *
     * This method is intended for derived scenario classes, when they need to set scenario methods of their own.
     */
    void reset_scenario_methods();

    template<typename S>
    inline void set_init(bool (S::*f)(int argc, char **argv))
    {
        invalidate();
        remove_init();
        if (f) _init = new scenario_init<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S>
    inline void set_finish(void (S::*f)())
    {
        invalidate();
        remove_finish();
        if (f) _finish = new scenario_finish<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S, typename T>
    inline void set_get_state(T (S::*f)())
    {
        invalidate();
        remove_get_state();
        if (f)
        {
            _get_state = new scenario_get_state<S,T>(*dynamic_cast<S*>(this), f);
        }
    }

    template<typename S>
    inline void set_get_state_id(String* (S::*f)(Object*))
    {
        invalidate();
        remove_get_state_id();
        if (f)
        {
            _get_state_id = new scenario_get_state_id<S>(*dynamic_cast<S*>(this), f);
        }
    }

    template<typename S>
    inline void set_is_stationary_state(bool (S::*f)())
    {
        invalidate();
        remove_is_stationary_state();
        if (f) _is_stationary = new scenario_is_stationary_state<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S>
    inline void set_observe_state(void (S::*f)())
    {
        invalidate();
        remove_observe_state();
        if (f) _observe_state = new scenario_observe_state<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S>
    inline void set_restore_model_state(void (S::*f)(Object*))
    {
        invalidate();
        remove_restore_model_state();
        if (f) _restore_state = new scenario_restore_model_state<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S, typename T>
    inline void set_save_model_state(T (S::*f)())
    {
        invalidate();
        remove_save_model_state();
        if (f) _save_model_state = new scenario_save_model_state<S>(*dynamic_cast<S*>(this), f);
    }

    template<typename S1, typename S2>
    void setup(const char * name,
                      bool (S1::*init)(int argc, char ** argv), void (S2::*finish)()
                     )
    {
        invalidate();
        _name = name;
        set_init(init);
        set_finish(finish);

        set_get_state(UseVirtual::get_state);
        set_get_state_id(UseVirtual::get_state_id);
        set_is_stationary_state(NoMethod::is_stationary_model_state);
        set_observe_state(NoMethod::observe_model_state);
        set_save_model_state(NoMethod::save_model_state);
        set_restore_model_state(NoMethod::restore_model_state);
    }

    template<typename S1, typename S2, typename T, typename S3>
    void setup(const char * name,
                bool (S1::*init)(int argc, char ** argv), void (S2::*finish)(),
                T (S3::*get_state)()
                     )
    {
        invalidate();
        _name = name;
        set_init(init);
        set_finish(finish);
        set_get_state(get_state);

        set_get_state_id(UseVirtual::get_state_id);
        set_is_stationary_state(NoMethod::is_stationary_model_state);
        set_observe_state(NoMethod::observe_model_state);
        set_save_model_state(NoMethod::save_model_state);
        set_restore_model_state(NoMethod::restore_model_state);
    }
    template<typename S1, typename S2, typename T, typename S3, typename S4>
    void setup(const char * name,
               bool (S1::*init)(int argc, char ** argv), void (S2::*finish)(),
               T (S3::*get_state)(),
               String* (S4::*get_state_id)(Object*)
               )
    {
        invalidate();
        _name = name;
        set_init(init);
        set_finish(finish);
        set_get_state(get_state);
        set_get_state_id(get_state_id);

        set_is_stationary_state(NoMethod::is_stationary_model_state);
        set_observe_state(NoMethod::observe_model_state);
        set_save_model_state(NoMethod::save_model_state);
        set_restore_model_state(NoMethod::restore_model_state);
    }
    template<typename S1, typename S2, typename T1, typename S3, typename S4,
             typename S5, typename S6, typename T2, typename S7>
    void setup(const char * name,
               bool (S1::*init)(int argc, char ** argv), void (S2::*finish)(),
               T1 (S3::*get_state)(),
               String* (S4::*get_state_id)(Object*),
               bool (S5::*is_stationary)(),
               T2 (S6::*save_model_state)(),
               void (S7::*restore_model_state)(Object*)
               )
    {
        invalidate();
        _name = name;
        set_init(init);
        set_finish(finish);
        set_get_state(get_state);
        set_get_state_id(get_state_id);
        set_is_stationary_state(is_stationary);
        set_save_model_state(save_model_state);
        set_restore_model_state(restore_model_state);
        set_observe_state(UseVirtual::observe_model_state);
    }

    template<typename S1, typename S2, typename T1, typename S3, typename S4,
             typename S5, typename S6, typename T2, typename S7, typename S8>
    void setup(const char * name,
                      bool (S1::*init)(int argc, char ** argv) , void (S2::*finish)() ,
                      T1 (S3::*get_state)() , String* (S4::*get_state_id)(Object*) ,
                      bool (S5::*is_stationary)() ,
                      T2 (S6::*save_model_state)() ,
                      void (S7::*restore_model_state)(Object*) ,
                      void (S8::*observe)()
                     )
    {
        invalidate();
        _name = name;
        set_init(init);
        set_finish(finish);
        set_get_state(get_state);
        set_get_state_id(get_state_id);
        set_is_stationary_state(is_stationary);
        set_save_model_state(save_model_state);
        set_restore_model_state(restore_model_state);
        set_observe_state(observe);
    }


    /// @brief Set all methods to default values.
    /// This method sets scenario name and intializes all methods with defaults.
    /// @param name const char * Scenario name.
    void setup(const char * name);
    virtual ~ScenarioBase();
public:

    virtual CppTestScenario * getTestScenario();
};

template<>
inline void ScenarioBase::set_init(bool (ScenarioBase::NoMethod::*f)(int argc, char **argv))
{
    remove_init();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_finish(void (ScenarioBase::NoMethod::*f)())
{
    remove_finish();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_get_state(Object* (ScenarioBase::NoMethod::*f)())
{
    remove_get_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_get_state_id(String* (ScenarioBase::NoMethod::*f)(Object*))
{
    remove_get_state_id();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_is_stationary_state(bool (ScenarioBase::NoMethod::*f)())
{
    remove_is_stationary_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_save_model_state(Object* (ScenarioBase::NoMethod::*f)())
{
    remove_save_model_state();
    // To avoid the warning
    if(f) {}
}
template<>
inline void ScenarioBase::set_observe_state(void (ScenarioBase::NoMethod::*f)())
{
    remove_observe_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_restore_model_state(void (ScenarioBase::NoMethod::*f)(Object*))
{
    remove_restore_model_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_init(bool (ScenarioBase::UseVirtual::*f)(int argc, char **argv))
{
    set_default_init();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_finish(void (ScenarioBase::UseVirtual::*f)())
{
    set_default_finish();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_get_state(Object* (ScenarioBase::UseVirtual::*f)())
{
    set_default_get_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_get_state_id(String* (ScenarioBase::UseVirtual::*f)(Object*))
{
    set_default_get_state_id();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_is_stationary_state(bool (ScenarioBase::UseVirtual::*f)())
{
    set_default_is_stationary_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_observe_state(void (ScenarioBase::UseVirtual::*f)())
{
    set_default_observe_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_save_model_state(Object* (ScenarioBase::UseVirtual::*f)())
{
    set_default_save_model_state();
    // To avoid the warning
    if(f) {}
}

template<>
inline void ScenarioBase::set_restore_model_state(void (ScenarioBase::UseVirtual::*f)(Object*))
{
    set_default_restore_model_state();
    // To avoid the warning
    if(f) {}
}

} // namespace cpptesk::ts
} // namespace cpptesk
