/** @file multififo_model.cpp
 * Model of united FIFOs.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <multififo_model.h>

namespace cpptesk {
namespace examples {
namespace fifo {

Multififo::Multififo() : iface1a("input iface1a multififo"),
	                 iface1b("input iface1b multififo"),
	                 iface2("input iface2 multififo"),
	                 iface3("output iface3 multififo"),
	                 fifo0(), fifo1(), fifo2()
{
    /// Registration of interfaces in the test system.
    CPPTESK_ADD_COMPONENT(iface1a);
    CPPTESK_ADD_COMPONENT(iface1b);
    CPPTESK_ADD_COMPONENT(iface2);
    CPPTESK_ADD_COMPONENT(iface3);
}

Multififo::~Multififo() {}

bool Multififo::isIface1aReady() const
{
    /// return true if iface1a is not busy by any operations.
    return iface1a.isReady();
}

bool Multififo::isIface1bReady() const
{
    /// return true if iface1b is not busy by any operations.
    return iface1b.isReady();
}

bool Multififo::isIface2Ready() const
{
    /// return true if iface2 is not busy by any operations.
    return iface2.isReady();
}

CPPTESK_DEFINE_STIMULUS(Multififo::push_msg_a)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Refresh model of Multififo.
    fifo0.push_item(data.get_data());
    fifo0.start(&FIFO::pop_msg, fifo0.iface2, data, FIFO::START_REQUEST_IFACE);

    // Set stimulus to the selected iface in parallel mode.
    START_STIMULUS(PARALLEL);

    // The process wait for 1 cycle.
    CYCLE();

    // Finish of stimulus processing.
    STOP_STIMULUS();
}

CPPTESK_DEFINE_STIMULUS(Multififo::push_msg_b)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Refresh model of Multififo.
    fifo1.push_item(data.get_data());
    fifo1.start(&FIFO::pop_msg, fifo1.iface2, data, FIFO::START_REQUEST_IFACE);

    // Set stimulus to the selected iface in parallel mode.
    START_STIMULUS(PARALLEL);

    // The process wait for 1 cycle.
    CYCLE();

    // Finish of stimulus processing.
    STOP_STIMULUS();
}

CPPTESK_DEFINE_STIMULUS(Multififo::pop_msg)
{
    // Copy data to local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Start of stimulus processing.
    START_PROCESS();

    /* Set stimulus ot the selected iface
       (the macro is used when one need to use non-standard iface or data).
     */
    RECV_STIMULUS(PARALLEL, iface, data);

    // Preparation of model reaction to check data given from DUT.
    OutputData outdata = OutputData(fifo2.pop_item());

    // Start reaction "take output data as a result of reading".
    get_pop_msg(process, iface3, outdata);

    // Finish of stimulus processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_REACTION(Multififo::get_pop_msg)
{
    START_PROCESS();
    // Registration of the reaction in mediator.
    SEND_REACTION(SEQUENTIAL, iface, message);

    STOP_PROCESS();
}

}}}
