/** @file fifo_media.cpp
 * Mediator of FIFO reference model.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <fifo_media.h>
#include <fifo_interface.h>

#include <iostream>

namespace cpptesk {
namespace examples {
namespace fifo {

FIFOMediator::FIFOMediator()
{
    // Registration of serializers, deserializers to interfaces.
    CPPTESK_SET_INPUT_ADAPTER(iface1, FIFOMediator::serialize_iface1);
    CPPTESK_SET_INPUT_ADAPTER(iface2, FIFOMediator::serialize_iface2);
    CPPTESK_SET_OUTPUT_ADAPTER(iface3, FIFOMediator::deserialize_iface3);

    // Setup of maximum time of waiting.
    setReactionTimeout(FIFOMediator::REACTION_TIMEOUT);

    // Initialization of structures to keep signals from DUT.
    clear_fifo_inputs(&inputs);
    clear_fifo_outputs(&outputs);

    // Registration of output interface iface3 arbiter.
    CPPTESK_SET_ARBITER(iface3, matcher_iface3);

    // Detector of unexpected reactions.
    CALL_OUTPUT_LISTENER(FIFOMediator::listen_iface3, iface3);
}

FIFOMediator::~FIFOMediator() {}

bool FIFOMediator::isIface1Ready() const
{
    return FIFO::isIface1Ready();
}
bool FIFOMediator::isIface2Ready() const
{
    return FIFO::isIface2Ready();
}

CPPTESK_DEFINE_STIMULUS(FIFOMediator::serialize_iface1)
{
    // Copy data into local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start working.
    START();

#ifdef DEBUG
    std::cout << "FIFOMediator::serialize_iface1" << std::endl;
#endif

    // Capture interface iface1.
    iface.capture();

    // Set data to wires of DUT.
    inputs.CE_WR = 1;
    inputs.DI = msg.get_data();

    // Release interface iface1.
    iface.release();

    // Wait for 1 cycle of simulation.
    CYCLE();

    // Stop processing.
    STOP();
}

CPPTESK_DEFINE_STIMULUS(FIFOMediator::serialize_iface2)
{
    // Copy data into local variable.
    InputData msg = CPPTESK_CAST_MESSAGE(InputData);

    // Start working.
    START();

#ifdef DEBUG
    std::cout << "FIFOMediator::serialize_iface2" << std::endl;
#endif

    // Capture interface iface2.
    iface.capture();

    // Set data to wires of DUT.
    inputs.CE_RD = 1;

    // Release interface iface2.
    iface.release();

    // Wait for 1 cycle of simulation.
    CYCLE();

    // Stop processing.
    STOP();
}

CPPTESK_DEFINE_REACTION(FIFOMediator::deserialize_iface3)
{
    // Copy data into local variable.
    OutputData &msg = CPPTESK_CAST_MESSAGE(OutputData);

    // Start working.
    START();

    // Wait for DO_VAL signal.
    WAIT_REACTION(outputs.DO_VAL);

    // Received data should be saved.
    msg.set_data(outputs.DO);

    // Jump to the following reaction if it is given.
    NEXT_REACTION();

    // Stop processing.
    STOP();
}
/*
#define PRINT_REACTION_INFO(iface, msg, info) \
{ \
    printf("-------------------------------------------------------------------------\n"); \
    printf("Reaction: cycle=%d\n", time()); \
    printf("Reaction: %s\n", iface.getName().c_str()); \
    printf("Reaction: %s\n", msg.toString().c_str()); \
    printf("Reaction: %s\n", info.empty() ? "[OK]" : info.c_str()); \
}
*/
/**
 * Function of reading from output ifaces.
 */
/*CPPTESK_DEFINE_PROCESS(FIFOMediator::onReaction)
{
    // Start default reaction processor.
    FIFO::onReaction(CPPTESK_GET_PARAMETERS());

    std::string info = error();

    if(!info.empty())
        { PRINT_REACTION_INFO(iface, message, info); }
}*/

void FIFOMediator::initialize()
{
//    ::start_testbench();
}

void FIFOMediator::finalize()
{
//    ::stop_testbench();
}

void FIFOMediator::setInputs()
{
    ::map_fifo_state_down(&inputs);
}

void FIFOMediator::getOutputs()
{
    ::map_fifo_state_up(&outputs);
}

void FIFOMediator::simulate()
{
}

}}}

