/** @file fifo_model.cpp
 * Reference Model of FIFO.
 *
 * @author Mikhail Chupilko (chupilko@ispras.ru)
 */

#include <fifo_model.h>

namespace cpptesk {
namespace examples {
namespace fifo {

FIFO::FIFO() : iface1("input iface1"), iface2("input iface2"), iface3("output iface3")
{
    /// Registration of interfaces in the test system.
    CPPTESK_ADD_COMPONENT(iface1);
    CPPTESK_ADD_COMPONENT(iface2);
    CPPTESK_ADD_COMPONENT(iface3);
}

FIFO::~FIFO() {}

bool FIFO::isIface1Ready() const
{
    /// Function isReady returns true, if iface1 is not busy by any operation.
    return iface1.isReady();
}

bool FIFO::isIface2Ready() const
{
    /// Function isReady returns true, if iface2 is not busy by any operation.
    return iface2.isReady();
}

CPPTESK_DEFINE_STIMULUS(FIFO::push_msg)
{
    // Copy data into local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Refresh of FIFO model.
    push_item(data.get_data());

    // Set stimulus to the received iface interface in parallel mode.
    START_STIMULUS(PARALLEL);

    // Wait for 1 cycle.
    CYCLE();

    // Stop processing.
    STOP_STIMULUS();
}

CPPTESK_DEFINE_STIMULUS(FIFO::pop_msg)
{
    // Copy data into local variable.
    InputData data = CPPTESK_CAST_MESSAGE(InputData);

    // Start processing.
    START_PROCESS();

    /* Set stimulus to the interface iface
       (this macro is used if one has not to use non-standard input interface of data name.
     */
    RECV_STIMULUS(PARALLEL, iface, data);

    // Preparation of model reaction to check data given by the DUT.
    OutputData outdata = OutputData(pop_item());

    // Start of reaction "Output data received in result of reading".
    get_pop_msg(process, iface3, outdata);

    // Stop processing.
    STOP_PROCESS();
}

CPPTESK_DEFINE_REACTION(FIFO::get_pop_msg)
{
    START_PROCESS();
    // Registration of the reaction in the mediator.
    SEND_REACTION(SEQUENTIAL, iface, message);

    STOP_PROCESS();
}

void FIFO::push_item(uint8_t data)
{
    // To continue loading buffer is prohibited if all items are used.
    assert(fifo.size() < 16);
    // Put data into the reference buffer.
    fifo.push_back(data);
}

uint8_t FIFO::pop_item(void)
{
    // To take data from buffer is impossible if they are absent.
    assert(fifo.size() > 0);
    // The data according to the FIFO strategy.
    uint8_t data = fifo[0];
    // Removing these data from buffer.
    fifo.erase(fifo.begin());
    // Return the data.
    return data;
}

bool FIFO::is_empty(void)
{
    return fifo.size() == 0;
}

bool FIFO::is_full(void)
{
    return fifo.size() == 16;
}

int FIFO::fifo_fulness(void)
{
    return fifo.size();
}

}}}

