/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Temporal sequence event
///////////////////////////////////////////////////////////////////////////////

#pragma once

#include <hw/defs.hpp>

#include <hw/clone.hpp>
#include <hw/unused.hpp>
#include <hw/utils/range.hpp>
#include <hw/utils/shared_ptr.hpp>

namespace cpptesk {
namespace hw {
namespace sequence {

////////////////////////////////////////////////////////////////////////////////
// Event
////////////////////////////////////////////////////////////////////////////////

/**
 * Event is a predicate being checked in a single point of time.
 */
class Event: public Cloneable
{
public:
    Event(): Cloneable() {}
    virtual ~Event() {}

    /// Clones the event object (sequence contains a copy of an event).
    Event *clone() const { return new Event(*this); } 

    /// The only interface method returns true iff the event is detected.
    bool match() const
    {
        if(occured())
        {
            handler();
            return true;
        }
        return false;
    }

    /// Checks if the event occured during the cycle.
    virtual bool occured() const { return true; }
    /// Handler is called when the event is detected.
    virtual void handler() const {}
};

/**
 * Abstract composition of two events.
 */
class CompositeEvent: public Event
{
public:
    CompositeEvent(const Event &lhs, const Event &rhs):
        Event(), lhs(lhs.clone()), rhs(rhs.clone()) {}
    virtual ~CompositeEvent() {}

    virtual bool occured() const = 0;
    virtual void handler() const
    {
        if(lhs->occured()) { lhs->handler(); }
        if(rhs->occured()) { rhs->handler(); }
    }

protected:
    SharedPtr<Event> lhs;
    SharedPtr<Event> rhs;
};

/**
 * Negation of an event: this->occured() = !sub->occured().
 */
class NotEvent: public Event
{
public:
    NotEvent(const Event &sub): Event(), sub(sub.clone()) {}
    virtual ~NotEvent() {}

    virtual bool occured() const { return !sub->occured(); }

protected:
    SharedPtr<Event> sub;
};

/**
 * Conjunction of two events: this->occured() = lhs->occured() && rhs->occured().
 */
class AndEvent: public CompositeEvent
{
public:
    AndEvent(const Event &lhs, const Event &rhs): CompositeEvent(lhs, rhs) {}
    virtual ~AndEvent() {}

    virtual bool occured() const { return lhs->occured() && rhs->occured(); }
};

/**
 * Disjunction of two events: this->occured() = lhs->occured() || rhs->occured().
 */
class OrEvent: public CompositeEvent
{
public:
    OrEvent(const Event &lhs, const Event &rhs): CompositeEvent(lhs, rhs) {}
    virtual ~OrEvent() {}
		
    virtual bool occured() const { return lhs->occured() || rhs->occured(); }

};

/**
 * Exclusive OR or two events: this->occured() = lhs->occured() ^ rhs->occured().
 */
class XorEvent: public CompositeEvent
{
public:
    XorEvent(const Event &lhs, const Event &rhs): CompositeEvent(lhs, rhs) {}
    virtual ~XorEvent() {}

    virtual bool occured() const { return lhs->occured() ^ rhs->occured(); }
};

////////////////////////////////////////////////////////////////////////////////
// Operations with events
////////////////////////////////////////////////////////////////////////////////

inline NotEvent operator ! (const Event &sub)
{
    return NotEvent(sub);
}

inline NotEvent operator ~ (const Event &sub)
{
    return NotEvent(sub);
}

inline AndEvent operator && (const Event &lhs, const Event &rhs)
{
    return AndEvent(lhs, rhs);
}

inline AndEvent operator & (const Event &lhs, const Event &rhs)
{
    return AndEvent(lhs, rhs);
}

inline OrEvent operator || (const Event &lhs, const Event &rhs)
{
    return OrEvent(lhs, rhs);
}

inline OrEvent operator | (const Event &lhs, const Event &rhs)
{
    return OrEvent(lhs, rhs);
}

inline XorEvent operator ^ (const Event &lhs, const Event &rhs)
{
    return XorEvent(lhs, rhs);
}

}}} // namespace cpptesk::hw::sequence

