/*
 * Copyright 2010-2011 ISP RAS (http://www.ispras.ru), UniTESK Lab (http://www.unitesk.com)
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *     http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

///////////////////////////////////////////////////////////////////////////////
// PROJECT:  C++TESK
// AUTHOR:   Alexander Kamkin (kamkin@ispras.ru)
// SYNOPSIS: Temporal sequence
///////////////////////////////////////////////////////////////////////////////

#include <hw/sequence/sequence.hpp>

namespace cpptesk {
namespace hw {
namespace sequence {

Sequence operator << (const Sequence &lhs, const Sequence &rhs)
{
    Sequence seq;

    // The left operand is a non-empty, single-repeated sequence
    if(lhs.isSequence())
    {
        seq = lhs;

        // The last element of the left sequence is not a delay
        if(!lhs.children.back().isEmpty())
        {
            // The both sequences are single-repeated
            if(rhs.isSequence())
            {
                // Merge the children
                seq.children.insert(seq.children.end(), rhs.children.begin(), rhs.children.end());
            }
            else
            {
                // Append the right sequence to the left's children
                seq.children.push_back(rhs);
            }
        }
        // The last element of the left sequence is a delay
        else
        {
            Delay delay = seq.children.back().delay;

            // The right sequence is single-repeated
            if(rhs.isSequence())
            {
                std::list<Sequence>::const_iterator rhs_children_second = rhs.children.begin();

                ++rhs_children_second;

                // Merge the children (remove the last left's delay and add it to the first right's delay)
                seq.children.erase(seq.children.end());
                seq.children.insert(seq.children.end(), rhs.children.begin(), rhs_children_second);
                seq.children.back().delay += delay;
                seq.children.insert(seq.children.end(), rhs_children_second, rhs.children.end());
            }
            else if(rhs.isSingleRepeated())
            {
                // Append the right sequence to the left's children
                seq.children.back() = rhs;
                seq.children.back().delay += delay;
            }
            else
            {
                // Append the right sequence to the left's children
                seq.children.push_back(rhs);
            }
        }
    }
    // The left operand is not a non-empty, single-repeated sequence, neither is the right one
    else if(rhs.isSequence())
    {
        seq = rhs;

        // The left sequence is a single-repeated delay
        if(lhs.isEmpty())
        {
            // Add the left's delay to the first right's delay
            seq.children.front().delay += lhs.delay;
        }
        else
        {
            // Append the left sequence to the right's children
            seq.children.push_front(lhs);
        }
    }
    // Both operands are not non-empty, single-repeated sequences
    else
    {
        // Both operands are single-repeated, and the left one is a delay
        if(lhs.isEmpty() && rhs.isSingleRepeated())
        {
            seq = rhs;
            seq.delay += lhs.delay;
        }
        else
        {
            // Create a new single-repeated sequence
            seq.children.push_back(lhs);
            seq.children.push_back(rhs);
        }
    }

    return seq;
}

////////////////////////////////////////////////////////////////////////////////
// Repetition of a sequence
////////////////////////////////////////////////////////////////////////////////

Sequence Sequence::repeat(unsigned min, unsigned max) const
{
    Sequence result;
    Repeat repetition(min, max);

    // OPTIMIZATION: S(1) = S
    if(repetition.isOne())
        { return *this; }

    // OPTIMIZATION: S(0) = ANY
    if(repetition.isZero())
        { return result; }

    // OPTIMIZATION: ANY(N) = ANY
    if(isEmpty() && delay.isZero())
        { return result; }

    // Nested REPEAT sequences
    if(isRepeat())
    {
        // OPTIMIZATION: S(N)(M) = S(N*M)
        if(this->repetition.isPoint() && repetition.isPoint())
        {
            result = *this;
            result.repetition.min = result.repetition.min * repetition.min;
            result.repetition.max = result.repetition.max * repetition.max;
            return result;
        }

        // OPTIMIZATION: S(K<=1,inf)(M<=1,N) = S(M,N)(K,inf) = S(K*M,inf)
        if((this->repetition.isStarOrPlus() && repetition.min <= 1) ||
           (repetition.isStarOrPlus() && this->repetition.min <= 1))
        {
            result = *this;
            result.repetition.min = result.repetition.min * repetition.min;
            result.repetition.max = Repeat::INFINITY;
            return result;
        }
    }

    return Sequence(*this, repetition);
}

////////////////////////////////////////////////////////////////////////////////
// Printing operator
////////////////////////////////////////////////////////////////////////////////
std::ostream& operator << (std::ostream &out, const Sequence &seq)
{
    if(seq.isSequence())
    {
        out << "(";
        for(std::list<Sequence>::const_iterator i = seq.children.begin(); i != seq.children.end(); i++)
            { out << (i != seq.children.begin() ? " " : "") << *i; }
        out << ")";
    }
    else
    {
        out << seq.delay << (!seq.isEmpty() ? " " : "");

        if(seq.isEvent())
            { out << "event"; }
        else if(seq.isGuard())
            { out << "wait_until(event)"; }
        else if(seq.isAnd() || seq.isOr())
            { out << "(" << seq.children.front() << (seq.isAnd() ? " && " : " || ") << seq.children.back() << ")"; }
        else if(seq.isRepeat())
            { out << seq.children.front() << seq.repetition; }
    }

    return out;
}

}}} // namespace cpptesk::hw::sequence
