#include <ts/system.h>
#include <atl/object_hashmap.h>
#include <netndfsm.h>
#include <synchronizer/synchronizer.h>

#include <stdio.h>


/* Define it if needed */
/* #define DEBUG_PRINT */


static int my_arcs[16];
static int my_index = 0;
static int iter_count;


static void conf_simple( Synchronizer *sync, int id )
{
	printf("Topology: 2 nodes\n");

	if (sync_SetInputConf( sync, (unsigned short)(12345 + id), 1) < 0)
		exit(1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short)(id==0 ? 12346 : 12345) );
}

static void conf_ring( Synchronizer *sync, int id )
{
	printf("Topology: unidirectional ring of 16 nodes\n");

	if (sync_SetInputConf( sync, (unsigned short)(12345 + id), 1 ) < 0)
		exit(1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short)(id==15 ? 12345 : 12345+id+1) );
}

static void conf_double_ring( Synchronizer *sync, int id )
{
	printf("Topology: bidirectional ring of 16 nodes\n");

	if (sync_SetInputConf( sync, (unsigned short)(12345 + id), 2) < 0)
		exit(1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short)(id==15 ? 12345 : 12345+id+1) );
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short)(id==0 ? 12345+15 : 12345+id-1) );
}

static void conf_torus( Synchronizer *sync, int id )
{
int hi, lo, port;

	printf("Topology: unidirectional torus of 16 nodes\n");
	
	hi = id/4;
	lo = id%4;
	if (sync_SetInputConf( sync, (unsigned short)(12345 + id), 2) < 0)
		exit(1);

	port = 12345 + (hi==3 ? 0 : hi+1) * 4 + lo;
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );
	port = 12345 + hi * 4 + (lo == 3 ? 0 : lo+1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );
}

static void conf_double_torus( Synchronizer *sync, int id )
{
int hi, lo, port;

	printf("Topology: bidirectional torus of 16 nodes\n");
	
	hi = id/4;
	lo = id%4;
	if (sync_SetInputConf( sync, (unsigned short)(12345 + id), 4) < 0)
		exit(1);

	port = 12345 + (hi==0 ? 3 : hi-1) * 4 + lo;
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );
	port = 12345 + hi * 4 + (lo == 0 ? 3 : lo-1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );

	port = 12345 + (hi==3 ? 0 : hi+1) * 4 + lo;
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );
	port = 12345 + hi * 4 + (lo == 3 ? 0 : lo+1);
	sync_AddOutConnection( sync, "127.0.0.1", (unsigned short) port );
}


static TSSemaphore db_sema = NULL;
static void generate_update();
static bool db_Update( bool local, TE_ArcId *arc_id );


void test_single_thread( Synchronizer *sync, int update_delay )
{
int i;
	for (i = 0; i < iter_count; i++) {
		generate_update();
		sync_DoExchange( sync );
		sleep_msec( update_delay );
	}
	sync_Shutdown( sync );
}

void test_multiple_threads( Synchronizer *sync, int exchange_delay, int update_delay )
{
int i;

	db_sema = ts_create_semaphore();

	sync_setExchangeDelay( sync, exchange_delay );
	sync_StartThread( sync );

	for (i = 0; i < iter_count; i++) {
		generate_update();
		sleep_msec( update_delay );
	}

	sync_FinishThread( sync );

	ts_destroy_semaphore( db_sema );
	db_sema = NULL;
}


int main(int argc, char **argv)
{
int i, id;
Synchronizer *sync;
TE_Graph graph;

	id = atoi(argv[1]) % 16;
	printf ("My id: %i\n", id);

	for (i = 0; i < /*15-*/id; i++) {
		my_arcs[i] = 1000 + 10*id + i;
	}
	my_arcs[/*15-*/id] = -1;

	init_Graph( &graph, TE_UPDATE_ON, &TE_DEFAULT_MUTEX );
	// init_TestEngine( &TE_NDFSM, &graph, some_test_here, argc, argv );

	sync = create_Synchronizer( &graph, &TE_NDFSM );

	// conf_simple(sync, id);
	// conf_ring(sync, id);
	// conf_double_ring(sync, id);
	// conf_torus(sync, id);
	conf_double_torus(sync, id);

	sleep_msec(2000);
	if ( sync_ConfigurationFinished( sync ) < 0 )
		exit(1);

	// iter_count = 10 + id;
	// iter_count = 16;
	// iter_count = 20;
	// iter_count = 70;
	iter_count = 200;

	// test_single_thread( sync, 20 );
	test_multiple_threads( sync, 1, 5 );

	destroy_Synchronizer( sync );

	return 0;
}

/** Emulation of test engine */
static void generate_update()
{
TE_ArcId *arc;
int i;

	i = my_arcs[my_index];
	if (i < 0)	return;

	arc = create_ArcId(
		create_NodeId( create_String("0") ),
		create_StimulusId(i),
		create_NodeId( create_String("0") )
	);
	arc->start_node_id->deg = 10;
	arc->end_node_id  ->deg = 10;

#ifdef DEBUG_PRINT
	{
		String *ss = toString( r(arc) );
		printf(" Test Engine update: %s\n", toCharArray_String(r(ss)));
		destroy(ss);
	}
#endif

	db_Update( true, arc );

	my_index++;
}

// DB stubs

static HashMap *db;
static TE_ArcList *local_updates;

static int hash_String(String *str)
{
const unsigned char *s = toCharArray_String( r(str) );
int hash = 0, l = length_String( r(str) );
int i;
	for (i = 0; i < l; i++) {
		hash = 31*hash + s[i];
	}
	destroy(str);
	return hash;
}

static int hash_TE_ArcId(TE_ArcId *arc)
{
int res = hash_String( r(arc->start_node_id->id) ) * 7
		+ arc->stimulus_id.index * 3
		+ hash_String( r(arc->end_node_id->id) );

	destroy(arc);
	return res;
}

TE_ArcList* getUpdates_TestEngine(TestEngine *engine, TE_Graph *graph)
{
TE_ArcList *res;

	if (db_sema != NULL)
		ts_close_semaphore(db_sema);

	if (local_updates != NULL) {
		res = local_updates;
		local_updates = NULL;
	}
	else res = create_List( &type_TE_ArcId );

	if (db_sema != NULL)
		ts_open_semaphore(db_sema);

	return res;
}

static bool db_Update( bool local, TE_ArcId *arc_id )
{
TE_UpdateStatus res = TE_OLD_ALL;

	if (db_sema != NULL)
		ts_close_semaphore(db_sema);

	if (db == NULL)
		db = create_HashMap( &type_TE_ArcId, &type_String, 1024, hash_TE_ArcId );

	if (!containsKey_HashMap( r(db), r(arc_id) )) {
		res = TE_NEW_ARC;
		put_HashMap( r(db), r(arc_id), NULL );
		if (local) {
			if (local_updates == NULL)
				local_updates = create_List( &type_TE_ArcId );
			append_List( r(local_updates), r(arc_id) );
		}
	}

	if (db_sema != NULL)
		ts_open_semaphore(db_sema);

	// no destroy() for arc_id
	return res;
}

TE_UpdateStatus updateLocal_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id, TE_NodeId *new_end_node_id)
{
	return db_Update(true,arc_id);
}

TE_UpdateStatus updateRemote_TestEngine(TestEngine *engine, TE_Graph *graph, TE_ArcId *arc_id)
{
	return db_Update(false,arc_id);
}
