/** @file ccp_model.cpp
 *
 * A model of the simple write-invalidate cache coherence protocol (CCP) used to
 * illustrate the C++TESK facilities on cache coherence (CC) verification using
 * the graph exploration (state enumeration) approach. The idea of the example
 * was borrowed from F. Pong and M. Dubois' paper "Verification Techniques for
 * Cache Coherence Protocols" (ACM Computing Surveys (CSUR), Volume 29 Issue 1,
 * March 1997).
 *
 * @author Alexander Kamkin (kamkin@ispras.ru)
 */

#include <ccp_model.h>

#include <algorithm>

namespace cpptesk {
namespace examples {
namespace ccp {

////////////////////////////////////////////////////////////////////////////////
// Cache
////////////////////////////////////////////////////////////////////////////////

#define UNEXPECTED() CPPTESK_ASSERTION(false, "Unexpected stimulus")

/**
 * Definition of the cache state machine.
 */
void Cache::process(requester_t r, operation_t o)
{
    switch(_state)
    {
    case INVALID:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { UNEXPECTED();     break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = INVALID; break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            break;
        }
    case SHARED:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { _state = SHARED;  break; }
                }
            }
            break;
        }
    case DIRTY:
        {
            if(r == THIS)
            {
                switch(o)
                {
                case READ:    { _state = DIRTY;   break; }
                case WRITE:   { _state = DIRTY;   break; }
                case REPLACE: { _state = INVALID; break; }
                }
            }
            else // OTHER
            {
                switch(o)
                {
                case READ:    { _state = SHARED;  break; }
                case WRITE:   { _state = INVALID; break; }
                case REPLACE: { UNEXPECTED();     break; }
                }
            }
            break;
        }
    default: { UNEXPECTED(); }
    }
}

////////////////////////////////////////////////////////////////////////////////
// Cache Coherence Model
////////////////////////////////////////////////////////////////////////////////

std::vector<std::string> CacheCoherenceModel::getStateVector() const
{
    std::vector<std::string> state;

    for(std::vector<Processor>::const_iterator i = _processors.begin(); i != _processors.end(); i++)
        { state.push_back(i->state()); }

    return state;
}

std::string CacheCoherenceModel::toString(const std::vector<std::string> &state) const
{
    std::stringstream buffer;

    for(std::vector<std::string>::const_iterator i = state.begin(); i != state.end(); i++)
        { buffer << *i; }

    return buffer.str();
}

/// Returns the true state.
std::string CacheCoherenceModel::getState() const
{
    return toString(getStateVector());
}

/// Exploits the sequence-to-multiset abstraction.
std::string CacheCoherenceModel::getMultiSetState() const
{
    std::vector<std::string> multiset = getStateVector();
    sort(multiset.begin(), multiset.end());
    return toString(multiset);
}

/// Exploits the sequence-to-set abstraction.
std::string CacheCoherenceModel::getSetState() const
{
    std::vector<std::string> set, seq = getStateVector();

    sort(seq.begin(), seq.end());

    for(std::vector<std::string>::const_iterator i = seq.begin(); i != seq.end();)
    {
        std::string item = *i;
        set.push_back(item);
        for(++i; i != seq.end() && (*i) == item; i++);
    }

    return toString(set);
}

}}} // namespace cpptesk::examples::ccp
